// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650
#ifndef FUNCTIONAL_H
#define FUNCTIONAL_H

#include <blitz/array.h>
#include "FIAT_polynomial.hpp"
#include "FIAT_rcp.hpp"

namespace FIAT {

  class Functional;
  class PointEvaluation;
  class FunctionalSet;

  class Functional
  {
  public:
    Functional( RCP<PolynomialBase> PB ,
		blitz::Array<double,1> dof ) : PB(PB) , dof(dof) {}
    double operator()( RCP<ScalarPolynomial> p ) const
    { return sum( dof * p->getDOF() ); }
    blitz::Array<double,1> getDOF( ) const { return dof; }

  private:
    RCP<PolynomialBase> PB;
    blitz::Array<double,1> dof;  
  };

  class PointEvaluation : public Functional
  {
  public:
    PointEvaluation( RCP<PolynomialBase> PB ,
		     blitz::Array<double,1> x ) :
      Functional( PB , PB->eval_all( x ) ) , x(x) {}
  private:
    blitz::Array<double,1> x;
  };


  class FunctionalSet
  {
  public:
    FunctionalSet( RCP<PolynomialBase> PB ,
		   vector<RCP<Functional> > ls ) :
      PB(PB) , mat( ls.size() , PB->size() ) 
    {
      for (int i=0;i<(int)ls.size();i++) {
	mat(i,blitz::Range::all()) = ls[i]->getDOF();
      }
    }

    int size() const { return mat.rows(); }
  
    Functional operator[]( int fnum ) const {
      if (fnum < 0 || fnum >= size()) FIAT_ERROR("invalid index " << fnum
                                                 << " in Functional operator[]");
      return Functional(PB,mat(fnum,blitz::Range::all())); 
    }


    blitz::Array<double,2> get_mat() const { return mat; }

    blitz::Array<double,2> outer_product( RCP<ScalarPolynomialSet> PS ) const
    {
      blitz::firstIndex i;
      blitz::secondIndex j;
      blitz::thirdIndex k;
      blitz::Array<double,2> PSmat = PS->get_mat();
      blitz::Array<double,2> result(mat.rows(),PSmat.columns());
      result = blitz::sum( mat(i,k) * PSmat(j,k) , k );
      return result;
    }

  private:
    RCP<PolynomialBase> PB;
    blitz::Array<double,2> mat;
  };
}
#endif
