// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650


#ifndef POLYNOMIAL_H
#define POLYNOMIAL_H

#include <map>
#include <blitz/array.h>
#include "FIAT_expansions.hpp"
#include "FIAT_shapes.hpp"
#include "FIAT_la.hpp"
#include "FIAT_rcp.hpp"

namespace FIAT {
  
  using namespace std;

  class PolynomialBase;
  class ScalarPolynomial;
  class ScalarPolynomialSet;
  class OrthonormalScalarPolynomialSet;

  blitz::Array<double,2> identity( int n );

  class PolynomialBase
  {

  public:
    PolynomialBase( const PolynomialBase & );
    PolynomialBase( int shp , int n );

    blitz::Array<double,1> eval_all( const blitz::Array<double,1> & ) const;
    blitz::Array<double,2> tabulate( const blitz::Array<double,2> & ) const;

    int degree() const { return n; }
    int domain_shape() const { return shp; }
    int size() const { return Shapes::polynomial_dimension( shp , n ); }
    int topological_dimension() const 
    { return Shapes::topological_dimension(shp); }
  
    blitz::Array<double,2> dmat( int dir ) const 
    { return dmats(dir,blitz::Range::all(),blitz::Range::all()); }
    
  private:
    const int shp;
    const int n;
    blitz::Array<double,3> dmats;
  };


  class ScalarPolynomial
  {
    friend class Functional;
  public:
    ScalarPolynomial( const ScalarPolynomial & );
    ScalarPolynomial( RCP<PolynomialBase> , blitz::Array<double,1> );
  
    double operator()( const blitz::Array<double,1> & );

    blitz::Array<double,1> getDOF() const { return dof; }

    RCP<ScalarPolynomial> deriv( int dir ) const;

  private:
    RCP<PolynomialBase> PB;
    blitz::Array<double,1> dof;
  };

  class ScalarPolynomialSet
  {
  public:
    ScalarPolynomialSet( RCP<PolynomialBase> ,
			 blitz::Array<double,2> );

    blitz::Array<double,1> eval_all( const blitz::Array<double,1> & ) const;
    blitz::Array<double,2> tabulate( const blitz::Array<double,2> & ) const;
    blitz::Array<double,2> trace_tabulate( int , int  , 
				    blitz::Array<double,2> ) const;

    int degree() const { return PB->degree(); }
    int domain_shape() const { return PB->domain_shape(); }
    int size() const { return coeffs.rows(); }
    int topological_dimension() const { return PB->topological_dimension(); }
    RCP<ScalarPolynomialSet> deriv_all( int dir ) const;
    RCP<ScalarPolynomialSet> multi_deriv_all( const blitz::Array<int,1> &alpha ) const;
    RCP<PolynomialBase> get_base() const { return PB; }
    RCP<ScalarPolynomial> operator[](int i);
    blitz::Array<double,2> get_mat() const { return coeffs; }

  protected:
    RCP<PolynomialBase> PB;
    blitz::Array<double,2> coeffs;
  };

  class OrthonormalScalarPolynomialSet : public ScalarPolynomialSet
  {
  public:
    OrthonormalScalarPolynomialSet( RCP<PolynomialBase> PB ) :
      ScalarPolynomialSet( PB , identity( PB->size() ) ) {}
				  
    OrthonormalScalarPolynomialSet( int shp , int n ) :
      ScalarPolynomialSet( new PolynomialBase(shp,n) ,
			   identity( Shapes::polynomial_dimension(shp,n) ) ) 
    {}
  };
}
#endif
