// Written by Robert C. Kirby
// Copyright 2005 by The University of Chicago
// Distributed under the LGPL license
// This work is partially supported by the US Department of Energy
// under award number DE-FG02-04ER25650

#ifndef SHAPES_H
#define SHAPES_H

#include <exception>
#include <cmath>
#include <blitz/array.h>

namespace FIAT {

  using namespace std;

  class Shapes;
  class PtMap;

  class Shapes
  {
  public:
    static const int LINE = 1;
    static const int TRIANGLE = 2;
    static const int TETRAHEDRON = 3;

    // returns the inherent topological dimension of the shp
    // (e.g. lines are one-dimensional)
    static int topological_dimension( const int shp );

    // number of entities of a particular dimension associated with
    // a given shape (e.g. tetrahedra have four vertices,
    // so num_entities(TETRAHEDRON,0) --> 4
    static int num_entities( int shp ,
			     int dimension );

    // inherent dimension of polynomials of degree n
    // on the given polynomial shape
    static int polynomial_dimension( int shp ,
				     int n );

    // gets coordinate of vertex number v_id associated with
    // the given shape
    static blitz::Array<double,1> get_vertex( int shp , int v_id );

    // constructs the lattice of equispaced points associated
    // with a given shape.  There are n+1 points on an edge
    // if interior == 1, then points on the boundary are omitted
    static blitz::Array<double,2> make_lattice( int shp ,
					 int n ,
					 int interior=0 );

    // returns the id's of the vertices associated with
    // a given shape for a given dimensional entity for
    // a given eid.
    static blitz::Array<int,1> get_vertex_ids( int shp , 
					int ent_dim ,
					int eid );

    // returns the points on component ent_id of
    // topological dimension ent_dim on the reference shape
    // embedded in lattice of order n
    static blitz::Array<double,2> make_points( int shp ,
					int ent_dim ,
					int ent_id ,
					int n );

    static blitz::Array<double,1> get_normal( int shp ,
				       int ent_id );

    static blitz::Array<double,1> get_tangent( int shp ,
					int ent_id );

    static PtMap &get_pt_map( int shp ,
			      int ent_dim ,
			      int ent_id );

  };

  //input must be correct size
  class PtMap
  {
  public:
    virtual void operator()(const blitz::Array<double,1> &,
			    blitz::Array<double,1> &) const = 0;
    virtual ~PtMap(){;}
  };

  class IdentityPtMap : public PtMap
  {
  public:
    void operator()(const blitz::Array<double,1> &xin,
		    blitz::Array<double,1> &xout) const 
    {
      xout = xin;
      return;
    }
    virtual ~IdentityPtMap(){;}
  };

  class PtToEdge : public PtMap
  {
  public:
    PtToEdge( int , const blitz::Array<int,1> & );
    void operator()(const blitz::Array<double,1> &,
		    blitz::Array<double,1> &) const;
    virtual ~PtToEdge(){;}
  private:
    const int shp;
    double a0,b0,h;
    blitz::Array<double,1> v0,v1,diff;
  };


  class PtToFaceTet : public PtMap
  {
  public:
    PtToFaceTet( const blitz::Array<int,1> & );
    void operator()(const blitz::Array<double,1> &xin ,
		    blitz::Array<double,1> &xout) const;
    virtual ~PtToFaceTet(){;}
  private:
    blitz::Array<double,1> v0,v1,v2,d0,d1;
  };
}
#endif
